import os
import json
import requests
from urllib.parse import urlparse
from multiprocessing.pool import ThreadPool
from tqdm import tqdm

from .utils import download_and_save_image, download_and_save_segmentation_bitmap, handle_exif_rotation

from PIL import Image

class SegmentsDataset():
    """SegmentsDataset class.

    Args:
        release_file (str or dict): Path to a release file, or a release dict resulting from client.get_release().
        labelset (str, optional): The labelset that should be loaded. Defaults to 'ground-truth'.
        filter_by (list, optional): A list of label statuses to filter by. Defaults to None.
        segments_dir (str, optional): The directory where the data will be downloaded to and stored. Defaults to 'segments'.

    """

    # https://stackoverflow.com/questions/682504/what-is-a-clean-pythonic-way-to-have-multiple-constructors-in-python
    def __init__(self, release_file, labelset='ground-truth', filter_by=None, segments_dir='segments'):
        self.labelset = labelset
        self.filter_by = [filter_by] if isinstance(filter_by, str) else filter_by
        if self.filter_by is not None:
            self.filter_by = [s.lower() for s in self.filter_by]
        self.segments_dir = segments_dir
        
        # if urlparse(release_file).scheme in ('http', 'https'): # If it's a url
        if isinstance(release_file, str): # If it's a file path
            with open(release_file) as f:
                self.release = json.load(f)
        else: # If it's a release object
            release_file = release_file['attributes']['url']
            content = requests.get(release_file)
            self.release = json.loads(content.content)        
        self.release_file = release_file

        self.dataset_identifier = '{}_{}'.format(self.release['dataset']['owner'], self.release['dataset']['name'])
        self.image_dir = os.path.join(segments_dir, self.dataset_identifier, self.release['name'])

        # First some checks
        if not self.labelset in self.release['dataset']['labelsets']:
            print('There is no labelset with name "{}".'.format(self.labelset))
            return

        self.task_type = self.release['dataset']['task_type']
        if self.task_type not in ['segmentation-bitmap', 'bboxes']:
            print('You can only create a dataset for tasks of type "segmentation-bitmap" or "bboxes" for now.')
            return
        
        self.load_dataset()

    def load_dataset(self):
        print('Initializing dataset. This may take a few seconds...')
        
        # Setup cache
        if not os.path.exists(self.image_dir):
            os.makedirs(self.image_dir)
        
        # Load and filter the samples
        samples = self.release['dataset']['samples']
        if self.filter_by is not None:
            filtered_samples = []
            for sample in samples:
                if sample['labels'][self.labelset] is not None:
                    label_status = sample['labels'][self.labelset]['label_status'].lower()
                else:
                    label_status = 'unlabeled'

                if label_status in self.filter_by:
                    filtered_samples.append(sample)
        else:
            filtered_samples = samples

        self.samples = filtered_samples
            
        # # Preload all samples (sequentially)
        # for i in tqdm(range(self.__len__())):
        #     item = self.__getitem__(i)

        # To avoid memory overflow or "Too many open files" error when using tqdm in combination with multiprocessing.
        def _load_image(i):
            self.__getitem__(i)
            return i

        # Preload all samples (in parallel)
        # https://stackoverflow.com/questions/16181121/a-very-simple-multithreading-parallel-url-fetching-without-queue/27986480
        # https://stackoverflow.com/questions/3530955/retrieve-multiple-urls-at-once-in-parallel
        # https://github.com/tqdm/tqdm/issues/484#issuecomment-461998250
        num_samples = self.__len__()
        with ThreadPool(16) as pool:
            # r = list(tqdm(pool.imap_unordered(self.__getitem__, range(num_samples)), total=num_samples))
            r = list(tqdm(pool.imap_unordered(_load_image, range(num_samples)), total=num_samples))

        print('Initialized dataset with {} images.'.format(num_samples))

        
    def _load_image_from_cache(self, sample):
        sample_name = os.path.splitext(sample['name'])[0]
        image_url = sample['attributes']['image']['url']
        url_extension = os.path.splitext(urlparse(image_url).path)[1]
        # image_filename_rel = '{}{}'.format(sample['uuid'], url_extension)
        image_filename_rel = '{}{}'.format(sample_name, url_extension)
        image_filename = os.path.join(self.image_dir, image_filename_rel)

        if not os.path.exists(image_filename):
            download_and_save_image(image_url, image_filename)

        image = Image.open(image_filename)
        image = handle_exif_rotation(image)

        return image, image_filename_rel

    def _load_segmentation_bitmap_from_cache(self, sample, labelset):
        sample_name = os.path.splitext(sample['name'])[0]
        label = sample['labels'][labelset]
        segmentation_bitmap_url = label['attributes']['segmentation_bitmap']['url']
        url_extension = os.path.splitext(urlparse(segmentation_bitmap_url).path)[1]
        # segmentation_bitmap_filename = os.path.join(self.image_dir, '{}{}'.format(label['uuid'], url_extension))
        segmentation_bitmap_filename = os.path.join(self.image_dir, '{}_label_{}{}'.format(sample_name, labelset, url_extension))
        
        if not os.path.exists(segmentation_bitmap_filename):
            download_and_save_segmentation_bitmap(segmentation_bitmap_url, segmentation_bitmap_filename)

        segmentation_bitmap = Image.open(segmentation_bitmap_filename)

        return segmentation_bitmap

    @property
    def categories(self):
        return self.release['dataset']['labelsets'][self.labelset]['attributes']['categories']
        # categories = {}
        # for category in self.release['dataset']['labelsets'][self.labelset]['attributes']['categories']:
        #     categories[category['id']] = category['name']
        # return categories

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, index):
        sample = self.samples[index]
        
        # Load the image
        try:
            image, image_filename = self._load_image_from_cache(sample)
        except:
            print('Something went wrong loading sample {}:'.format(sample['name']), sample)
            raise

        item = {
            'uuid': sample['uuid'],
            'name': sample['name'],
            'file_name': image_filename,
            'image': image,
        }

        # Segmentation bitmap
        if self.task_type == 'segmentation-bitmap':
            # Load the label
            try:
                label = sample['labels'][self.labelset]
                segmentation_bitmap = self._load_segmentation_bitmap_from_cache(sample, self.labelset)
                attributes = label['attributes']
                annotations = attributes['annotations']
            except:
                segmentation_bitmap = attributes = annotations = None
            
            item.update({
                'segmentation_bitmap': segmentation_bitmap,
                'annotations': annotations,
                'attributes': attributes
            })

        # Bounding boxes
        elif self.task_type == 'bboxes':
            try:
                label = sample['labels'][self.labelset]
                attributes = label['attributes']
                annotations = attributes['annotations']
            except:
                attributes = annotations = None

            item.update({
                'annotations': annotations,
                'attributes': attributes
            })

        else:
            assert False

#         # transform
#         if self.transform is not None:
#             item = self.transform(item)

        return item


import os
from io import BytesIO
import requests
import urllib.request
import json
from tqdm import tqdm

from PIL import Image, ExifTags

import numpy as np


def bitmap2file(bitmap, is_segmentation_bitmap=True):
    """Convert a label bitmap to a file with the proper format.

    Args:
        bitmap (np.uint32): A numpy array where each unique value represents an instance id.

    Returns:
        object: a file object.
    """

    # TODO: convert bitmap to np.uint32, if it is not already
    if is_segmentation_bitmap:
        bitmap2 = np.copy(bitmap)
        bitmap2 = bitmap2[:, :, None].view(np.uint8)
        bitmap2[:, :, 3] = 255
    else:
        assert False
        
    f = BytesIO()
    Image.fromarray(bitmap2).save(f, 'PNG')
    f.seek(0)
    return f

def get_semantic_bitmap(instance_bitmap, annotations, first_category_id=1):
    """Convert an instance bitmap and annotations dict into a segmentation bitmap.

    Args:
        instance_bitmap (np.uint32): A numpy array where each unique value represents an instance id.
        annotations (dict): An annotations dictionary.
        first_category_id (int, optional): First category id. Defaults to 1.

    Returns:
        np.uint32: a numpy array where each unique value represents a category id.
    """

    instance2semantic = [0] * (max([a['id'] for a in annotations])+1)
    for annotation in annotations:
        instance2semantic[annotation['id']] = annotation['category_id'] + first_category_id
    instance2semantic = np.array(instance2semantic)
        
    semantic_label = instance2semantic[np.array(instance_bitmap, np.uint32)]
    return semantic_label

def export_dataset(dataset, export_format='coco'):
    """Export a dataset to a different format.

    Args:
        dataset (dict): A dataset object, resulting from client.get_dataset().
        export_format (str, optional): The destination format. Defaults to 'coco'.
    """

    from pycocotools import mask
    from skimage.measure import regionprops

    def get_bbox(binary_mask):
        regions = regionprops(np.uint8(binary_mask))
        if len(regions) == 1:
            bbox = regions[0].bbox
            return bbox
        else:
    #         assert False
            return False

    # https://www.immersivelimit.com/tutorials/create-coco-annotations-from-scratch/#coco-dataset-format

    if export_format != 'coco':
        print('Supported export formats: coco')
        return

    info = {
        'description': dataset.release['dataset']['name'],
        # 'url': 'https://segments.ai/test/test',
        'version':  dataset.release['name'],
        # 'year': 2020,
        # 'contributor': 'Segments.ai',
        # 'date_created': '2020/09/01'
    }

    # licenses = [{
    #     'url': 'http://creativecommons.org/licenses/by-nc-sa/2.0/',
    #     'id': 1,
    #     'name': 'Attribution-NonCommercial-ShareAlike License'
    # }]

    categories = dataset.categories
    task_type = dataset.task_type
    # for i, category in enumerate(dataset.project_info['label_taxonomy']):
    #     categories.append({
    #         'id': i+1,
    #         'supercategory': 'object',
    #         'name': category
    #     })

    images = []
    annotations = []

    annotation_id = 1
    for i in tqdm(range(len(dataset))):        
        sample = dataset[i]
    #     print(sample)
        
        image_id = i+1
        images.append({        
            'id': image_id,
            # 'license': 1,
            'file_name': sample['file_name'],
            'height': sample['image'].size[1],
            'width': sample['image'].size[0],
    #         'date_captured': "2013-11-14 17:02:52",
    #         'coco_url': "http://images.cocodataset.org/val2017/000000397133.jpg",
    #         'flickr_url': "http://farm7.staticflickr.com/6116/6255196340_da26cf2c9e_z.jpg",        
        })
        
        for instance in sample['annotations']:
            category_id = instance['category_id']

            annotation = {
                'id': annotation_id,
                'image_id': image_id,
                'category_id': category_id,
            }

            # Segmentation bitmap
            if task_type == 'segmentation-bitmap':                
                instance_mask = np.array(sample['segmentation_bitmap'], np.uint32) == instance['id']
                bbox = get_bbox(instance_mask)
                if not bbox:
                    continue
                    
                y0, x0, y1, x1 = bbox
                # rle = mask.encode(np.asfortranarray(instance_mask))
                rle = mask.encode(np.array(instance_mask[:,:,None], dtype=np.uint8, order='F'))[0] # https://github.com/matterport/Mask_RCNN/issues/387#issuecomment-522671380
        #         instance_mask_crop = instance_mask[y0:y1, x0:x1]
        #         rle = mask.encode(np.asfortranarray(instance_mask_crop))
        #         plt.imshow(instance_mask_crop)
        #         plt.show()
                
                area = int(mask.area(rle))
                rle['counts'] = rle['counts'].decode('ascii')

                annotation.update({
                    'bbox': [x0, y0, x1-x0, y1-y0],
        #             'bbox_mode': BoxMode.XYWH_ABS,
                    'segmentation': rle,
                    'area': area,
                    'iscrowd': 0,
                })

            # Bounding boxes
            elif task_type == 'bboxes':
                points = instance['points']
                x0 = points[0][0]
                y0 = points[0][1]
                x1 = points[1][0]
                y1 = points[1][1]

                annotation.update({
                    'bbox': [x0, y0, x1-x0, y1-y0],
                })

            else:
                assert False

            annotations.append(annotation)
            annotation_id += 1
            
    json_data = {
        'info': info,
        # 'licenses': licenses,
        'categories': categories,
        'images': images,
        'annotations': annotations    
    #     'segment_info': [] # Only in Panoptic annotations
    }

    file_name = '{}_coco.json'.format(os.path.splitext(os.path.basename(dataset.release_file))[0])
    with open(file_name, 'w') as f:
        json.dump(json_data, f)

    print('Exported to {}.'.format(file_name))
    return file_name, dataset.image_dir

def download_and_save_image(url, filename):
    urllib.request.urlretrieve(url, filename)
    return

def download_and_save_segmentation_bitmap(url, filename):
    def extract_segmentation_bitmap(segmentation_bitmap):
        segmentation_bitmap = np.array(segmentation_bitmap)
        segmentation_bitmap[:,:,3] = 0
        segmentation_bitmap = segmentation_bitmap.view(np.uint32).squeeze(2)
        segmentation_bitmap = Image.fromarray(segmentation_bitmap)
        return segmentation_bitmap

    segmentation_bitmap = Image.open(BytesIO(requests.get(url).content))
    segmentation_bitmap = extract_segmentation_bitmap(segmentation_bitmap)
    segmentation_bitmap.save(filename)
    return

def handle_exif_rotation(image):
    try:
        for orientation in ExifTags.TAGS.keys():
            if ExifTags.TAGS[orientation]=='Orientation':
                break
        exif=dict(image._getexif().items())
        if exif[orientation] == 3:
            image=image.transpose(Image.ROTATE_180)
        elif exif[orientation] == 6:
            image=image.transpose(Image.ROTATE_270)
        elif exif[orientation] == 8:
            image=image.transpose(Image.ROTATE_90)
        return image
    except (AttributeError, KeyError, IndexError):
        return image